#!/usr/bin/perl -w

# shared_ngram_collector.pl
# (was duplicate_spotter_mysql_XX.pl
#
# version 0.41 (Powerslave)
#
# by Marco Baroni & Eros Zanchetta
# 
# The WaCky Project
# http://wacky.sslmit.unibo.it/

# for documentation and usage try
# shared_ngram_collector.pl -h | more

use strict;
use Getopt::Std;
use Digest::MD4 qw(md4 md4_base64);	# the fastest fingerprinting,
					# according to documentation
use DBI;
use Time::HiRes qw( gettimeofday tv_interval );
use String::Random;	# we use this to generate a random name for the DB, it
			# has no other uses

my $usage;
{
$usage = <<"_USAGE_";

This script takes as input a document collection, and it outpus a set
of files useful in the search for (near) duplicates in the collection.

The script is our implementation of a variation of the famous
shingling algorithm proposed in:

Andrei Z. Broder, Steven C. Glassman, Mark S. Manasse and Geoffrey
Zweig (1997). Syntactic Clustering of the Web. Sixth International
World-Wide Web Conference.

The basic idea is that if two documents share a certain number of long
n-grams, then the two documents are similar. For reasons of
efficiency, rather than comparing all the n-grams in each document,
the script samples a fixed number of random n-grams from each
document, and it produces a list of n-grams and the documents that
share them. From this list, it is straightforward to measure the
degree of overlap between two documents (we have post-processing tools
we use internally and we could make public: if you are interested in
them please email us).

Notice that we sample from n-gram types, not tokens, i.e., each
distinct n-gram occurring in a document has the same probability of
being sampled, no matter what its frequency inside the document is,
and the same n-gram cannot be sampled more than once from the same
document.

Moreover, the script normalizes the input before collecting n-grams,
lower-casing characters in the latin1 range, removing most
non-alphanumericals, etc.

The user can specify the size of the sample from each document with
the option -s and the length of the n-grams with the option -n. By
default, the script samples 20 pentagrams from each document (i.e., -s
and -n default to 20 and 5, respectively).

The user can also pass a stop list with the option -w (for
"unigrams"), one word per line. In this case, the script will remove
the stop words before collecting unigrams (e.g., if "the" is in the
stop list, "the good the bad the ugly" becomes "good bad ugly").

Moreover, the user can specify a list of words that occur only once in
the corpus or (better) only in one document. This list of hapax
legomena (passed via the option -l) is likely to make the script much
more efficient (the implementation exploits the fact that if a n-gram
contains a hapax, we can be sure from the offset that it will not be
shared by two or more documents).

The script can also be used as a tokenizer/normalizer by passing the
option -t.  The reason for this is that, if you want to provide a list
of stop words and/or hapax legomena, you should make sure that tokens
are segmented and normalized exactly as the script would. The easiest
way to do this is to use the script itself to tokenize and normalize
the input as it would normally do, and then extract the relevant list
from the output of the tokenization (btw: the script does a pretty
dumb job at tokenizing/normalizing the input: it s not meant to be
used as general purpose tokenizer).

The input collection must be in a single file. Each document in the
collection must be prefixed by a line that contains a non-ambiguous
delimiter string followed by optional space and a string (not
containing space) uniquely identifying the document. For example, the
delimiter could be "CURRENT URL" and the ids could be urls, in which
case the delimitng lines should look like:

CURRENT URL http://www.blahblahblah.org

The delimiter can be specified with the option -d; default is "CURRENT
DOC".

The script generates the following output files (where "prefix" is a
string that can be specified via option -x, or "duplicate.spotter" by
default):

- prefix.docs: list of the urls of all documents along with their ids

- prefix.dup: list of perfect duplicates.

- prefix.ngrams: list of all ngrams with their MD4 sums

- prefix.sim: is a list of shared ngrams and all the ids of all the 
  documents that share them.

- prefix.small: lists documents that are too "small", in the sense
  that, after tokenization, normalization and stop word removal, they do
  not contain at least s n-grams (e.g., if sample size is set to 20 and
  n to 5, this file will contain the list of the document ids of
  documents that did not contain at least 20 distinct pentagrams after
  tokenization); for our purposes (creation of corpora from web-data),
  very short documents are probably not too interesting anyway.)

The script requires the following perl modules to be installed (some
should be part of default installation):

Digest::MD4
DBI
Time::HiRes
String::Random

Moreover, the script needs to talk to a MySQL database server.

Usage:

$0 [-d "delim"] [-x pref] [-w stop.list]
     [-l hapax.list] [-s X] [-n N] doc_collection
                                  [> stdout.msgs]

$0 [-d "DELIMITER STRING"] -t doc_collection
                              > tokenized_corpus

$0 -h | more

Options:

-d delimiter: the string marking lines delimiting documents
   (default: "CURR DOC")

-l hapax.list: file with hapax legomena, one per line

-m hostname: MySQL database server (default: localhost)

-n n: the number of elements n in a n-gram (default: 5)

-p password: MySQL password (default is no password)

-r port: MySQL port (default: 3306)

-s sample: the number of n-grams to sample from each document
   (default: 20)

-u username: MySQL username (default: root)

-w stop.list: file with stop words, one per line

-x prefix: prefix string for output files (default: duplicate.spotter)

Switches:

-c: print a few statistics

-h: print this information and quit

-k: don't remove temporary database when finished

-t: instead of looking for shared n-grams, print tokenized/normalized
    version of input corpus (for stop/hapax list extraction purposes)

-v: print version number and quit


Copyright 2005 Marco Baroni, Eros Zanchetta, WaCky Project

This program is free software. You may copy or redistribute it under the same
terms as Perl itself.

_USAGE_
}
{
	my $blah = 1;
	# this useless block is here because here document confuses
	# emacs
}

my %opts = ();
# getopts('d:u:n:s:p:thl:',\%opts);
getopts('d:l:m:n:p:r:s:u:w:x:chktv',\%opts);

if ($opts{v}) {
	print "shared_ngram_collector.pl\nVersion 0.4\n";
	exit;
}

if ($opts{h}) {
	print $usage;
	exit;
}

if ($opts{t}) {
	# tokenization is not compatible with other options, die
	if ($opts{w} || $opts{n} || $opts{s} || $opts{x}) {
		die "when tokenizing, only delimiter option can be specified\n";
	}
}

my %stop_list;
if ($opts{w}) {
	open STOPLIST,$opts{w} or die "could not open stop list file $opts{w}";
	while (<STOPLIST>) {
		chomp;
		$stop_list{$_} = 1;
	}
	close STOPLIST;
}

my %hapax_list;
if ($opts{l}) {
	open HAPAXLIST,$opts{l} or die "could not open hapax legomena list file $opts{l}";
	while (<HAPAXLIST>) {
		chomp;
		$hapax_list{$_} = 1;
	}
	close HAPAXLIST;
}

# default delimiter: CURRENT DOC
my $delimiter;
if (!($delimiter = $opts{d})) {
	$delimiter = "CURRENT DOC";
}

# default: pentagrams
my $n;
if (!($n = $opts{n})) {
	$n = 5;
}

# how many of them will we pick? by default 20
# (s stands for subset, I guess)
my $s;
if (!($s = $opts{s})) {
	$s = 20;
}

# prefix for output files
my $pref;
if (!($pref = $opts{x})) {
	$pref = "duplicate\.spotter";
}

die "Oops! You must specify an input file.\n"
	if (!$ARGV[0]);

# if not tokenizing,
# check that corresponding files do not already exist
if (!$opts{t}) {
	die "Oops! ${pref}.small already exists.\n"
		if -e "$pref\.small";
	die "Oops! ${pref}.dup already exists.\n"
		if -e "$pref\.dup";
	die "Oops! ${pref}.sim already exists.\n"
		if -e "$pref\.sim";
	die "Oops! ${pref}.docs already exists.\n"
		if -e "$pref\.docs";
	die "Oops! ${pref}.ngrams already exists.\n"
		if -e "$pref\.ngrams";
}

# following variables must be initialized here, or else
# when in tokenization mode script will die
my $database_name;
my $db_handle;
my $doc_names;
my $doc_names_doc_names;
my $doc_names_ids;
my $lookup_name;
my $lookup_name_ngrams;

# if we are not in tokenization only mode
if (!$opts{t}) {
# DB initialization and relevant global variables

my $db_rand = new String::Random;

$database_name = $db_rand ->randpattern("cccccccc");  # generates an eight-character long lowercase string

my $host_name;
if (!($host_name = $opts{m})) {
	$host_name = "127.0.0.1";
}

my $port_number;
if (!($port_number = $opts{r})) {
	$port_number = "3306";
}

my $user_name;
if (!($user_name = $opts{u})) {
	$user_name = "root";
}

my $db_password;
if (!($db_password = $opts{p})) {
	$db_password = "";
}

# lookup table definition
$lookup_name = "lookup";
$lookup_name_ngrams = "ngrams";
my $lookup_name_doc_ids = "doc_ids";

# doc_names table definition
$doc_names = "doc_names";
$doc_names_ids = "ids"; 
$doc_names_doc_names = "doc_names";

# connect to database server
print "Opening connection to MySQL database server $host_name\n";
my $create_db_handle =
	DBI->connect("dbi:mysql:host=$host_name:port=$port_number;user=$user_name;password=$db_password")
	or die "Couldn't connect to database server: $DBI::errstr\n";

# create database
print "Creating temporary database $database_name\n";
my $create_db_query = "CREATE DATABASE $database_name";

$create_db_handle -> do($create_db_query) or die "Can't create db\n";

# connect to newly created db
print "Connecting to database $database_name\n";
$db_handle = DBI->connect("dbi:mysql:database=$database_name;host=$host_name:port=$port_number;user=$user_name;password=$db_password")
or die "Couldn't connect to database: $DBI::errstr\n";

# create document index
print "Creating tables\n";
my $create_doc_names_query =	"CREATE TABLE
					`$doc_names` (
						`$doc_names_ids` int(11) unsigned not NULL auto_increment,
						`$doc_names_doc_names` text NOT NULL,
						PRIMARY KEY (`$doc_names_ids`)
					)
					ENGINE=MyISAM DEFAULT
					CHARSET=latin1;";

$db_handle -> do($create_doc_names_query) or die "Can't create table $doc_names\n";
	
# create lookup table
my $create_lookup_query =		"CREATE TABLE
						`$lookup_name` (
							`$lookup_name_ngrams` char(22) not NULL default '',
							`$lookup_name_doc_ids` int(11) unsigned NOT NULL default '0',
							PRIMARY KEY (`$lookup_name_ngrams`,`$lookup_name_doc_ids`)
		  				)
		  				ENGINE=MyISAM DEFAULT
						CHARSET=latin1;";

$db_handle -> do($create_lookup_query) or die "Can't create table $lookup_name\n";
}

# some global variables
my @tokens = ();
my $currdoc;
my $newdoc;
my %digestlist = ();
#my %shared = ();

# if we are not tokenizing,
# we want to print to 4 files:
if (!$opts{t}){
	# one with files that are too small
	open SMALL,">$pref.small";
	
	# one with docs that turn out to be identical to something else
	open TRUEDUPS,">$pref.dup";
	
	# one with doc ids of docs that share n-grams
	open SIM,">$pref.sim";
	
	# one with doc - id correspondence
	open DOCS,">$pref.docs";
	
	# one with a dump of all md4 - ngrams pairs including repeated ones for further processing
	open NGRAMS, ">$pref.ngrams";
}

my $ts0 = gettimeofday();

while (<>) {

	# if we are at delimiter, it's time to process tokens array
	# and clean up

	if (/^[ \t]*$delimiter[ \t]*([^ \t]+)/) {
		$newdoc = $1;
		chomp $newdoc;
	
		# if this is not first doc, process data
		if ($currdoc) {
			
			# tokenization: just print
			if ($opts{t}) {
				# we need at least one token, don't we?
				if ($tokens[0]) {
					print "$delimiter $currdoc\n";
					print join "\n",@tokens;
					print "\n";
				}
			}
			# else, process data
			else {
				process_tokens_and_ngrams($currdoc,\@tokens);
			}
		}
	
		@tokens = ();
		$currdoc = $newdoc;
	}

	# if not delimiter, keep cleaning and tokenizing data

	else {
		chomp;
		# probably always a good idea to undos
		s/\r//g;
		
		# tokenize
		
		# the approach used here (clean, split at spaces, re-clean
		# tokens generated in this way) seems to be more efficient
		# than the character-based approach (split the line into
		# characters and then build tokens from characters)
			
		# substitute evil ms apostrophes
		s/\x92/\x27/g;
		
		# ... and evil mysterious space
		s/\xa0/ /g;
		
		# sequences of punctuation marks immediately preceding spaces
		# (or end of line) can be removed, since they are likely to be
		# true punctuation marks, not part of the token that precedes
		s/[\.\,\!\)\]\}\:\;\?\>\'\"\_\*]+($|[\s]+)/ /g;
		
		# and the same goes for some word-initial non-alphanumeric symbols
		s/(^|[\s]+)[\[\(\<\{\'\"\xbf\xa1\_\*]+/ /g;
		
		# also, if there is more than one - it's probably some kind of
		# boundary marker
		s/\-\-+/ /g;
		
		# ok, now besides dashes, apostrphes and slashes we should
		# have only internal non-alhpanumericals left which mark
		# tokens we probably want to discard; we keep the other
		# tokens, turning them to lower case; we also discard words in
		# stop word list, if any
	
		push @tokens, grep {/^[\x41-\x5a\x61-\x7a\xc0-\xd6\xd8-\xf6\xf8-\xff]([\x41-\x5a\x61-\x7a\xc0-\xd6\xd8-\xf6\xf8-\xff\-\'\/]*[\x41-\x5a\x61-\x7a\xc0-\xd6\xd8-\xf6\xf8-\xff]$|$)/ && !$stop_list{$_}}
					map &latin1_lc($_), split "[ \t\r]+",$_;
	}
}
# process last file

# my ($ts0, $tm0) = gettimeofday;
my $ts1 = gettimeofday();

if ($opts{t}) {
	# we need at least one token, don't we?
	if ($tokens[0]) {
		print "$delimiter $currdoc\n";
		print join "\n",@tokens;
		print "\n";
	}
}
else {
	process_tokens_and_ngrams($currdoc,\@tokens);
}

# unless we are in tokenization mode, its' time to
# print ngrams and docs that share them and cleanup
if (!$opts{t}) {
	# print ngrams and list of docs that share them
	# only for docs shared
	
	# my ($ts1, $tm1) = gettimeofday;
	my $ts2 = gettimeofday();
	
	printout_shared_ngrams();
	
	# my ($ts2, $tm2) = gettimeofday;
	my $ts3 = gettimeofday();
	
	printout_doc_names();
	
	# my ($ts3, $tm3) = gettimeofday;
	my $ts4 = gettimeofday();
	
	# a few statistics
	if ($opts{c}) {
		print "Finished preprocessing after: " . ($ts1 - $ts0) . " seconds\n";
		print "Finished populating DB after: " . ($ts2 - $ts1) . " seconds\n";
		print "Finished printing shared ngrams after: " . ($ts3 - $ts2) . " seconds\n";
		print "Finished printing document names after: " . ($ts4 - $ts4) . " seconds\n";
		print "Totale execution time: " . ($ts4 - $ts0) . " seconds\n";
	}
	
	# if user doesn't wish to keep it (default value) remove database
	
	if (!$opts{k}) {
		print "Removing temporary database $database_name\n";
		my $destroy_db_query = "DROP database $database_name;";  
		$db_handle -> do($destroy_db_query ) or die "Unable to remove database\n";
	}
	else {
		print "Keeping temporary database $database_name as specified by user\n";
	}
	
	# close the output files
		close SMALL;
		close TRUEDUPS;
		close SIM;
		close DOCS;
		close NGRAMS;
	
	print "Finished processing\n";
}


sub process_tokens_and_ngrams {
	my $currdoc = shift;
	my $tokensref = shift;
	
	
	# tokens must be at least s + n -1
	# it's true that I check this again later, but otoh
	# this preliminary check should be fast and helpful in
	# immediately discarding very small files
	if (@{$tokensref} < ($s+$n-1)) {
		# print docs that are too small to SMALL file
		print SMALL "$currdoc\t",scalar(@{$tokensref}),"\n";
	}
	else {
		# first, check if we already have perfect
		# (post-cleaning) duplicate: we print to file with
		# fingerprints anyway, but we keep processing only if
		# file does not have duplicate
		
		my $digest = md4(@{$tokensref}); # we lose white space info but
		# for reasonably long token strings
		# only truly strange freak accidents
		# would lead to sequences that
		# become identical when whitespaces
		# are removed
		
		my $dupdoc;
		# if we saw the digest already, let's print doc name to
		# true dup file, together with name of first doc that
		# had same digest
		if ($dupdoc = $digestlist{$digest}) {
			print TRUEDUPS "$currdoc\t$dupdoc\n";
		}
		# else, we record new digest, and move on with
		# processing
		else {
			$digestlist{$digest} = $currdoc;
			
			# randomly select s ngrams
	
			# unfortunately, since we do not want a bias in
			# favor of frequent ngrams, we will have to go
			# through list and extract ngrams (OK, I
			# could probably do something smarter in which I
			# still directly point to the token list, but for
			# now I want to get the full procedure done, I
			# will worry about optimization later... perhaps...)
			
			my %seen = ();
			my @ngrams = ();
	
			# keeping only one ngram token per ngram type
			foreach (0..(@{$tokensref}-$n)) {
				my $ngram = join " ",@{$tokensref}[$_..($_+$n-1)];
				if (!$seen{$ngram}) {
					$seen{$ngram} = 1;
					push @ngrams, $ngram;
				}
			}

			# could it be that now we are below the minimum
			# ngram number?
	
			if ($#ngrams < ($s-1)) {
				# print docs that are too small to SMALL file
				print SMALL "$currdoc\t",scalar(@{$tokensref}),"\n";
			}
	
			else {
				# finally, the good stuff -- we randomly select a
				# subset of s n-grams
				
				# let's register a doc id in the db
				my $currdoc_id = insert_doc_in_db_and_return_id($currdoc);
		
				my %already_generated = ();
				my $added = 0;
				
				srand();
				
				my @temp_ngrams_list = ();
				while ($added < $s) {
					$added++;
	
					my $ranindex = int (rand(@ngrams));
					
					while ($already_generated{$ranindex}) {
						$ranindex =  int (rand(@ngrams));
					}
					$already_generated{$ranindex}++;
		
					#push @{$shared{$ngrams[$ranindex]}}, $currdoc;
					my $curr_ngram = $ngrams[$ranindex];
	
					# if we got a list of hapaxes and a ngram
					# contains a hapax, there is not meaning
					# in keeping it since it could not possibly
					# be shared by more than one document
					if ($opts{l}) {
						my @components = split " ",$curr_ngram;
						my $found_hapax = 0;
						foreach (@components) {
							if ($hapax_list{$_}) {
								$found_hapax = 1;
								last;
							}
						}
						if ($found_hapax) {
							next;
						}
					}
					my $temp_digest = md4_base64($curr_ngram);
					push @temp_ngrams_list, $temp_digest;
						
					print NGRAMS $temp_digest . "\t" . $curr_ngram . "\n";
						
				}
				# we need at least one ngram!
				if (defined($temp_ngrams_list[0])) {
					add_to_lookup_id(\@temp_ngrams_list, $currdoc_id);
				}
			}
		}
	}
}
	

sub  insert_doc_in_db_and_return_id {
	my $currdoc = shift;
	
	$currdoc = $db_handle -> quote($currdoc);
	my $query_insert_doc_in_db = "INSERT into $doc_names VALUES ('', $currdoc);";
	$db_handle -> do($query_insert_doc_in_db) or die "Cannot insert doc in database\nFailed query:\n$query_insert_doc_in_db\n";
	
	my $query_find_last = "SELECT LAST_INSERT_ID();";
	my $statement = $db_handle->prepare($query_find_last);
	
	$statement->execute();
	my $row = $statement->fetchrow();
	return $row;
}


sub add_to_lookup_id {
	
	my $ngram_digest_array_ref = shift;
	my $currdoc_id = shift;
	
	# my $ngram_digest = $db_handle -> quote($ngram_digest);
	my $insert_record_query = "INSERT INTO `$lookup_name` VALUES ";
	
	my $virgola = "";
	foreach my $temp_ngram (@{$ngram_digest_array_ref}) {
		my $temp_ngram = $db_handle -> quote($temp_ngram);
		$insert_record_query .= $virgola . "(" . $temp_ngram . ", " . $currdoc_id . ")";
		$virgola = ",";
	}
	
	$db_handle -> do($insert_record_query) or die "Can't insert record\nFailed query:\n$insert_record_query\n";
}

sub printout_shared_ngrams {

	my $find_shared_ngrams = "
				SELECT a.*
					FROM $lookup_name as a
						INNER JOIN (
							SELECT
								$lookup_name_ngrams
							FROM
								$lookup_name
							GROUP BY
								$lookup_name_ngrams having count($lookup_name_ngrams) > 1
						) as b
		  			ON
  					a.$lookup_name_ngrams = b.$lookup_name_ngrams;";
	
	my $statement = $db_handle->prepare($find_shared_ngrams);
	
	$statement->execute();
	my $current_ngram = "";
	my $spacer = "";
	my @row = ();
	while (@row = $statement->fetchrow_array) {
		if ($row[0] ne $current_ngram) {
			if ($current_ngram ne "") {
				print SIM "\n";
			};
			$current_ngram = $row[0];
			print SIM $current_ngram . " ";
			$spacer = "";
		}
		print SIM $spacer;
		print SIM $row[1];
		$spacer = " ";
	}
	# add a newline at the end of the file, but only if the file is not empty
	if ($statement->rows > 1) { print SIM "\n"; }
}

sub printout_doc_names {
	my $query = "SELECT $doc_names_ids, $doc_names_doc_names FROM $doc_names;";
	my $statement = $db_handle->prepare($query);
	
	$statement->execute();
	while (my @row = $statement->fetchrow_array()) {
		print DOCS join ("\t", @row), "\n";
	}
}

sub latin1_lc {
	
	return join '', map {
		if ( ( (ord($_) > 64) && (ord($_) < 91) ) ||
		( (ord($_) > 192) && (ord($_) < 215) ) ||
		( (ord($_) > 215) && (ord($_) < 224) ) ) {
			sprintf "%c",(ord($_)+32);
		}
		else {
			sprintf "%s",$_;
		}
	}
	split '',shift;
}
